﻿using System.Collections.Generic;
using System.IO;
using System.Threading.Tasks;

namespace Inedo.Agents.Hosting
{
    public sealed class HostedProcess
    {
        private HashSet<int> sessions = new HashSet<int>();

        internal HostedProcess(SlimProcess process, AgentType agentType, int? id, SharedMemoryMessenger messenger, string binaryPath)
        {
            this.Process = process;
            this.AgentType = agentType;
            this.Messenger = messenger;
            this.BinaryPath = binaryPath;
            if (id != null)
                this.sessions.Add((int)id);

            this.ProcessIncomingMessages();
        }

        public AgentType AgentType { get; }
        public int? ProcessId
        {
            get
            {
                try { return this.Process.Id; }
                catch { return null; }
            }
        }
        private SharedMemoryMessenger Messenger { get; }
        public string FileName { get; }

        internal SlimProcess Process { get; }
        internal string BinaryPath { get; }
        internal bool Terminated { get; private set; }

        internal bool HasSession(int sessionId)
        {
            lock (this.sessions)
            {
                return this.sessions.Contains(sessionId);
            }
        }
        internal void AddSession(int sessionId)
        {
            lock (this.sessions)
            {
                this.sessions.Add(sessionId);
            }
        }
        internal void RemoveSession(int sessionId)
        {
            lock (this.sessions)
            {
                this.sessions.Remove(sessionId);
                if (this.sessions.Count == 0 && this.Terminated)
                    this.TerminateProcessAsync();
            }
        }
        internal void Terminate(bool force)
        {
            lock (this.sessions)
            {
                this.Terminated = true;
                if (force || this.sessions.Count == 0)
                {
                    this.sessions.Clear();
                    this.TerminateProcessAsync();
                }
            }
        }

        internal Task PostMessageAsync(int sessionId, AgentMessage message)
        {
            var rubbish = new MemoryStream((int)message.MessageData.Length + 4);
            var writer = new BinaryWriter(rubbish);
            writer.Write(sessionId);
            writer.Write(message.MessageTypeCode);
            message.MessageData.CopyTo(rubbish);
            rubbish.Position = 0;

            return this.Messenger.SendMessageAsync(rubbish);
        }

        private async void ProcessIncomingMessages()
        {
            await Task.Yield();
            while (true)
            {
                var stream = await this.Messenger.GetNextMessageAsync();
                if (stream == null)
                    return;

                var reader = new BinaryReader(stream);
                int sessionId = reader.ReadInt32();
                int messageId = reader.ReadInt32();

                HostedProcessManager.Instance.OnResponseReceived(new CommandResponseEventArgs(sessionId, messageId, this, new OffsetStream(stream, 8)));
            }
        }
        private async void TerminateProcessAsync()
        {
            try { this.Messenger.Dispose(); }
            catch { }

            await Task.Delay(5000);

            int processId = this.ProcessId ?? 0;
            if (processId > 0)
            {
                try
                {
                    using (var process = System.Diagnostics.Process.GetProcessById(processId))
                    {
                        process.Kill();
                    }
                }
                catch
                {
                }
            }
        }
    }
}
