﻿using System;
using System.Collections.Generic;
using System.IO;
using System.IO.Compression;
using System.Linq;
using System.Text.RegularExpressions;
using System.Threading.Tasks;

namespace Inedo.Agents.Hosting
{
    internal sealed class VirtualAgentHost
    {
        private static readonly Regex NameRegex = new Regex("^bin(?<1>[0-9]*)$", RegexOptions.Compiled | RegexOptions.CultureInvariant | RegexOptions.IgnoreCase | RegexOptions.ExplicitCapture);
        private List<HostedProcess> active = new List<HostedProcess>();
        private HostedProcess current;

        public VirtualAgentHost(AgentType agentType, ScopedConfig config)
        {
            this.AgentType = agentType;
            this.Config = config;
            this.CleanStaleAgentsAsync();
        }

        public static string MessengerPrefix { get; } = "Inedo_";
        public AgentType AgentType { get; }
        public ScopedConfig Config { get; }

        public void Start()
        {
            lock (this.active)
            {
                this.GetOrStartProcess(null);
            }
        }
        public void Stop(bool force)
        {
            lock (this.active)
            {
                if (force)
                {
                    foreach (var p in this.active)
                        p.Terminate(true);
                }

                this.current = null;
            }
        }
        public void Restart()
        {
            lock (this.active)
            {
                foreach (var p in this.active)
                    p.Terminate(false);

                this.current = null;
                this.GetOrStartProcess(null);
            }
        }

        public Task SendMessageAsync(int sessionId, AgentMessage message)
        {
            HostedProcess hostedProcess;

            lock (this.active)
            {
                hostedProcess = this.GetOrStartProcess(sessionId);
                if (hostedProcess == null)
                    throw new AgentUnavailableException(message.AgentType);
            }

            return hostedProcess.PostMessageAsync(sessionId, message);
        }
        public void DisconnectSession(int sessionId)
        {
            lock (this.active)
            {
                foreach (var process in this.active)
                    process.RemoveSession(sessionId);
            }
        }

        internal void InstallAgentUpdate(Commands.UpdateAgentCommand update)
        {
            lock (this.active)
            {
                if (update.Assemblies != null && update.Assemblies.Length > 0)
                {
                    var rootPath = this.Config.BinariesPath;
                    Directory.CreateDirectory(rootPath);

                    var existingNames = (from d in Directory.EnumerateDirectories(rootPath, "bin*", SearchOption.TopDirectoryOnly)
                                         let m = NameRegex.Match(Path.GetFileName(d))
                                         where m.Success
                                         let n = m.Groups[1]?.Value
                                         select !string.IsNullOrEmpty(n) ? (int?)int.Parse(n) : null).ToList();

                    var name = "bin";
                    if (existingNames.Contains(null))
                        name = "bin" + ((existingNames.Max() ?? 0) + 1);

                    var binPath = Path.Combine(rootPath, name);
                    Directory.CreateDirectory(binPath);

                    foreach (var asm in update.Assemblies)
                    {
                        var fileName = Path.Combine(binPath, asm.Name);
                        using (var targetStream = File.Create(fileName))
                        using (var sourceStream = new GZipStream(asm.Data, CompressionMode.Decompress))
                        {
                            sourceStream.CopyTo(targetStream);
                        }

                        File.SetLastWriteTimeUtc(fileName, asm.LastWriteTime);
                    }

                    this.SetCurrentAgentName(name);
                }

                if (update.Extensions != null && update.Extensions.Length > 0)
                {
                    var extensionsPath = this.Config.ExtensionsPath;
                    Directory.CreateDirectory(extensionsPath);

                    foreach (var extension in update.Extensions)
                    {
                        var fileName = Path.Combine(extensionsPath, extension.Name);
                        if (extension.Data == null || extension.Data.Length == 0)
                        {
                            File.Delete(fileName);
                        }
                        else
                        {
                            using (var source = extension.Data)
                            using (var fileStream = File.Create(fileName))
                            {
                                source.CopyTo(fileStream);
                            }

                            File.SetLastWriteTimeUtc(fileName, extension.LastWriteTime);
                        }
                    }
                }

                foreach (var p in this.active)
                    p.Terminate(false);

                this.current = null;
                this.GetOrStartProcess(null);
            }
        }

        private HostedProcess GetOrStartProcess(int? sessionId)
        {
            var hostedProcess = this.current;

            if (sessionId != null)
            {
                int id = (int)sessionId;
                foreach (var process in this.active)
                {
                    if (process.HasSession(id))
                    {
                        hostedProcess = process;
                        break;
                    }
                }
            }

            if (hostedProcess == null)
            {
                var currentName = this.GetCurrentAgentName();
                if (!string.IsNullOrEmpty(currentName))
                {
                    var path = Path.Combine(this.Config.BinariesPath, currentName, this.AgentType.ToString() + ".Agent.exe");
                    if (!File.Exists(path))
                        return null;

                    var messengerId = MessengerPrefix + Guid.NewGuid().ToString("N");
                    var messenger = SharedMemoryMessenger.Create(messengerId, SharedMemoryMessenger.DefaultBufferSize);

                    try
                    {
                        hostedProcess = this.StartProcess(path, this.AgentType, sessionId, currentName);
                    }
                    catch
                    {
                        try { messenger.Dispose(); }
                        catch { }
                        return null;
                    }
                }

                if (this.current == null)
                    this.current = hostedProcess;
            }

            if (hostedProcess != null && sessionId != null)
                hostedProcess.AddSession((int)sessionId);

            return hostedProcess;
        }

        private HostedProcess StartProcess(string fileName, AgentType agentType, int? sessionId, string binaryPath)
        {
            var messengerId = MessengerPrefix + Guid.NewGuid().ToString("N");
            var messenger = SharedMemoryMessenger.Create(messengerId, SharedMemoryMessenger.DefaultBufferSize);

            try
            {
                SlimProcess process = null;
                try
                {
                    process = SlimProcess.Start(fileName, messengerId, this.HandleProcessExited, this.Config);
                    var hostedProcess = new HostedProcess(process, agentType, sessionId, messenger, binaryPath);
                    active.Add(hostedProcess);
                    return hostedProcess;
                }
                catch
                {
                    process?.Dispose();
                    throw;
                }
            }
            catch
            {
                messenger?.Dispose();
                throw;
            }
        }

        private void HandleProcessExited(SlimProcess process)
        {
            lock (this.active)
            {
                this.active.RemoveAll(p => p.Process == process);
                if (this.current?.Process == process)
                    this.current = null;
            }
        }

        private string GetCurrentAgentName()
        {
            var fileName = Path.Combine(this.Config.BinariesPath, ".current");
            if (File.Exists(fileName))
            {
                try
                {
                    return File.ReadAllText(fileName);
                }
                catch (FileNotFoundException)
                {
                }
            }

            return null;
        }
        private void SetCurrentAgentName(string name)
        {
            var fileName = Path.Combine(this.Config.BinariesPath, ".current");
            try
            {
                File.WriteAllText(fileName, name);
            }
            catch
            {
            }
        }

        private async void CleanStaleAgentsAsync()
        {
            await Task.Yield();

            while (true)
            {
                if (!string.IsNullOrEmpty(this.Config.BinariesPath))
                {
                    try
                    {
                        var current = this.GetCurrentAgentName();
                        var directories = (from d in Directory.EnumerateDirectories(this.Config.BinariesPath)
                                           where NameRegex.IsMatch(Path.GetFileName(d)) && !string.Equals(current, Path.GetFileName(d), StringComparison.OrdinalIgnoreCase)
                                           select d).ToList();

                        foreach (var directory in directories)
                        {
                            var name = Path.GetFileName(directory);
                            bool inclused = false;
                            lock (this.active)
                            {
                                foreach (var p in this.active)
                                {
                                    if (string.Equals(p.BinaryPath, name, StringComparison.OrdinalIgnoreCase))
                                    {
                                        inclused = true;
                                        break;
                                    }
                                }

                                if (!inclused)
                                {
                                    try { Directory.Delete(directory, true); }
                                    catch { }
                                }
                            }
                        }
                    }
                    catch
                    {
                    }
                }

                await Task.Delay(new TimeSpan(0, 30, 0));
            }
        }
    }
}
